import 'package:dio/dio.dart';

import 'http_cancel_token.dart';
import 'http_error.dart';
import 'http_options.dart';

typedef HttpParams = Map<String, dynamic>;
typedef HttpRunnerWrapper<T> = Future<T?> Function({
  HttpParams? params,
  HttpCancelToken? cancelToken,
  HttpOptions? options,
  ProgressCallback? onSendProgress,
  ProgressCallback? onReceiveProgress,
});
typedef HttpRunner<T> = Future<T> Function({
  HttpParams? params,
  HttpCancelToken? cancelToken,
  HttpOptions? options,
  ProgressCallback? onSendProgress,
  ProgressCallback? onReceiveProgress,
});

/// Create API functions with perfect error handling.
///
/// ```
/// var http = Http.getInstances(HttpBaseOptions({
///   name: 'defaultApi',
///   baseUrl: 'http://domain.com/api/v1',
/// }));
///
/// class UserApi {
///   static final getUser = buildHttpRunnerWrapper<UserEntity>(({
///     HttpParams? params,
///     HttpCancelToken? cancelToken,
///     HttpOptions? options,
///     ProgressCallback? onSendProgress,
///     ProgressCallback? onReceiveProgress,
///   }) async {
///     var res = await http.get('/user/${params?['id']??1}',
///       queryParameters: {
///         'getFullData': params?['getFullData'] ?? '0',
///       },
///       cancelToken: cancelToken);
///
///     return UserEntity.fromJson(res.data);
///   });
/// }
///
/// try {
///   var user = await UserApi.getUser(params: {'id': 123, 'getFullData': 1});
///   if (user == null) {
///     print('Request cancelled!');
///   }
/// } on HttpError catch (e) {
///   if (e.errorCode == HttpErrorCode.notFound) {
///     print('The user could not be found');
///   }
/// }
/// ```
HttpRunnerWrapper<T> buildHttpRunnerWrapper<T>(HttpRunner<T> runner) {
  return ({
    HttpParams? params,
    HttpCancelToken? cancelToken,
    HttpOptions? options,
    ProgressCallback? onSendProgress,
    ProgressCallback? onReceiveProgress,
  }) async {
    try {
      return await runner(
        params: params,
        cancelToken: cancelToken,
        options: options,
        onSendProgress: onSendProgress,
        onReceiveProgress: onReceiveProgress,
      );
    } on HttpError catch (e) {
      throw e;
    } on DioError catch (e) {
      if (e.type == DioErrorType.cancel) {
        return null;
      }

      int statusCode = e.response?.statusCode ?? 0;
      throw HttpError(
        requestOptions: e.requestOptions,
        response: e.response,
        type: e.type,
        error: e.error,
        statusCode: statusCode,
        errorCode: HttpErrorCode.unknownError,
        errorMessage: null,
        errorData: e.response?.data,
      );
    }
  };
}
